@@ -417,7 +417,7 @@ format_get_number_of_directives (void *descr)
 
 static bool
 format_check (void *msgid_descr, void *msgstr_descr, bool equality,
-              formatstring_error_logger_t error_logger,
+              formatstring_error_logger_t error_logger, void *error_logger_data,
               const char *pretty_msgid, const char *pretty_msgstr)
 {
   struct spec *spec1 = (struct spec *) msgid_descr;
@@ -427,14 +427,16 @@ format_check (void *msgid_descr, void *msgstr_descr, bool equality,
   if (spec1->named_arg_count > 0 && spec2->unnamed_arg_count > 0)
     {
       if (error_logger)
-        error_logger (_("format specifications in '%s' expect a mapping, those in '%s' expect a tuple"),
+        error_logger (error_logger_data,
+                      _("format specifications in '%s' expect a mapping, those in '%s' expect a tuple"),
                       pretty_msgid, pretty_msgstr);
       err = true;
     }
   else if (spec1->unnamed_arg_count > 0 && spec2->named_arg_count > 0)
     {
       if (error_logger)
-        error_logger (_("format specifications in '%s' expect a tuple, those in '%s' expect a mapping"),
+        error_logger (error_logger_data,
+                      _("format specifications in '%s' expect a tuple, those in '%s' expect a mapping"),
                       pretty_msgid, pretty_msgstr);
       err = true;
     }
@@ -457,7 +459,8 @@ format_check (void *msgid_descr, void *msgstr_descr, bool equality,
               if (cmp > 0)
                 {
                   if (error_logger)
-                    error_logger (_("a format specification for argument '%s', as in '%s', doesn't exist in '%s'"),
+                    error_logger (error_logger_data,
+                                  _("a format specification for argument '%s', as in '%s', doesn't exist in '%s'"),
                                   spec2->named[j].name, pretty_msgstr,
                                   pretty_msgid);
                   err = true;
@@ -468,7 +471,8 @@ format_check (void *msgid_descr, void *msgstr_descr, bool equality,
                   if (equality)
                     {
                       if (error_logger)
-                        error_logger (_("a format specification for argument '%s' doesn't exist in '%s'"),
+                        error_logger (error_logger_data,
+                                      _("a format specification for argument '%s' doesn't exist in '%s'"),
                                       spec1->named[i].name, pretty_msgstr);
                       err = true;
                       break;
@@ -491,7 +495,8 @@ format_check (void *msgid_descr, void *msgstr_descr, bool equality,
                                   || spec2->named[j].type == FAT_ANY))))
                       {
                         if (error_logger)
-                          error_logger (_("format specifications in '%s' and '%s' for argument '%s' are not the same"),
+                          error_logger (error_logger_data,
+                                        _("format specifications in '%s' and '%s' for argument '%s' are not the same"),
                                         pretty_msgid, pretty_msgstr,
                                         spec2->named[j].name);
                         err = true;
@@ -512,7 +517,8 @@ format_check (void *msgid_descr, void *msgstr_descr, bool equality,
           if (spec1->unnamed_arg_count != spec2->unnamed_arg_count)
             {
               if (error_logger)
-                error_logger (_("number of format specifications in '%s' and '%s' does not match"),
+                error_logger (error_logger_data,
+                              _("number of format specifications in '%s' and '%s' does not match"),
                               pretty_msgid, pretty_msgstr);
               err = true;
             }
@@ -524,7 +530,8 @@ format_check (void *msgid_descr, void *msgstr_descr, bool equality,
                             || spec2->unnamed[i].type == FAT_ANY))))
                 {
                   if (error_logger)
-                    error_logger (_("format specifications in '%s' and '%s' for argument %u are not the same"),
+                    error_logger (error_logger_data,
+                                  _("format specifications in '%s' and '%s' for argument %u are not the same"),
                                   pretty_msgid, pretty_msgstr, i + 1);
                   err = true;
                 }
