@@ -1,6 +1,5 @@
 /* Checking of messages in PO files.
-   Copyright (C) 1995-1998, 2000-2008, 2010-2015 Free Software
-   Foundation, Inc.
+   Copyright (C) 1995-2023 Free Software Foundation, Inc.
    Written by Ulrich Drepper <drepper@gnu.ai.mit.edu>, April 1995.
 
    This program is free software: you can redistribute it and/or modify
@@ -14,7 +13,7 @@
    GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
-   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */
+   along with this program.  If not, see <https://www.gnu.org/licenses/>.  */
 
 #ifdef HAVE_CONFIG_H
 # include <config.h>
@@ -24,8 +23,6 @@
 #include "msgl-check.h"
 
 #include <limits.h>
-#include <setjmp.h>
-#include <signal.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -41,10 +38,6 @@
 #include "plural-table.h"
 #include "c-strstr.h"
 #include "message.h"
-#include "quote.h"
-#include "sentence.h"
-#include "unictype.h"
-#include "unistr.h"
 #include "gettext.h"
 
 #define _(str) gettext (str)
@@ -70,21 +63,15 @@ plural_expression_histogram (const struct plural_distribution *self,
       unsigned long n;
       unsigned int count;
 
-      /* Protect against arithmetic exceptions.  */
-      install_sigfpe_handler ();
-
       count = 0;
       for (n = min; n <= max; n++)
         {
-          unsigned long val = plural_eval (expr, n);
+          struct eval_result res = plural_eval (expr, n);
 
-          if (val == j)
+          if (res.status == PE_OK && res.value == j)
             count++;
         }
 
-      /* End of protection against arithmetic exceptions.  */
-      uninstall_sigfpe_handler ();
-
       return count;
     }
   else
@@ -101,12 +88,13 @@ int
 check_plural_eval (const struct expression *plural_expr,
                    unsigned long nplurals_value,
                    const message_ty *header,
-                   struct plural_distribution *distribution)
+                   struct plural_distribution *distribution,
+                   xerror_handler_ty xeh)
 {
   /* Do as if the plural formula assumes a value N infinitely often if it
      assumes it at least 5 times.  */
 #define OFTEN 5
-  unsigned char * volatile array;
+  unsigned char *array;
 
   /* Allocate a distribution array.  */
   if (nplurals_value <= 100)
@@ -115,99 +103,69 @@ check_plural_eval (const struct expression *plural_expr,
     /* nplurals_value is nonsense.  Don't risk an out-of-memory.  */
     array = NULL;
 
-  if (sigsetjmp (sigfpe_exit, 1) == 0)
+  unsigned long n;
+
+  for (n = 0; n <= 1000; n++)
     {
-      unsigned long n;
-
-      /* Protect against arithmetic exceptions.  */
-      install_sigfpe_handler ();
-
-      for (n = 0; n <= 1000; n++)
+      struct eval_result res = plural_eval (plural_expr, n);
+      if (res.status != PE_OK)
         {
-          unsigned long val = plural_eval (plural_expr, n);
+          if (res.status == PE_INTDIV)
+            xeh->xerror (CAT_SEVERITY_ERROR, header, NULL, 0, 0, false,
+                         _("plural expression can produce division by zero"));
+          else if (res.status == PE_INTOVF)
+            xeh->xerror (CAT_SEVERITY_ERROR, header, NULL, 0, 0, false,
+                         _("plural expression can produce integer overflow"));
+          else if (res.status == PE_STACKOVF)
+            xeh->xerror (CAT_SEVERITY_ERROR, header, NULL, 0, 0, false,
+                         _("plural expression can produce stack overflow"));
+          else
+            /* Other res.status values should not occur.  */
+            abort ();
 
-          if ((long) val < 0)
-            {
-              /* End of protection against arithmetic exceptions.  */
-              uninstall_sigfpe_handler ();
-
-              po_xerror (PO_SEVERITY_ERROR, header, NULL, 0, 0, false,
-                         _("plural expression can produce negative values"));
-              free (array);
-              return 1;
-            }
-          else if (val >= nplurals_value)
-            {
-              char *msg;
-
-              /* End of protection against arithmetic exceptions.  */
-              uninstall_sigfpe_handler ();
-
-              msg = xasprintf (_("nplurals = %lu but plural expression can produce values as large as %lu"),
-                               nplurals_value, val);
-              po_xerror (PO_SEVERITY_ERROR, header, NULL, 0, 0, false, msg);
-              free (msg);
-              free (array);
-              return 1;
-            }
-
-          if (array != NULL && array[val] < OFTEN)
-            array[val]++;
+          free (array);
+          return 1;
         }
 
-      /* End of protection against arithmetic exceptions.  */
-      uninstall_sigfpe_handler ();
+      unsigned long val = res.value;
 
-      /* Normalize the array[val] statistics.  */
-      if (array != NULL)
+      if ((long) val < 0)
         {
-          unsigned long val;
-
-          for (val = 0; val < nplurals_value; val++)
-            array[val] = (array[val] == OFTEN ? 1 : 0);
+          xeh->xerror (CAT_SEVERITY_ERROR, header, NULL, 0, 0, false,
+                       _("plural expression can produce negative values"));
+          free (array);
+          return 1;
+        }
+      else if (val >= nplurals_value)
+        {
+          char *msg =
+            xasprintf (_("nplurals = %lu but plural expression can produce values as large as %lu"),
+                       nplurals_value, val);
+          xeh->xerror (CAT_SEVERITY_ERROR, header, NULL, 0, 0, false, msg);
+          free (msg);
+          free (array);
+          return 1;
         }
 
-      distribution->expr = plural_expr;
-      distribution->often = array;
-      distribution->often_length = (array != NULL ? nplurals_value : 0);
-      distribution->histogram = plural_expression_histogram;
-
-      return 0;
+      if (array != NULL && array[val] < OFTEN)
+        array[val]++;
     }
-  else
+
+  /* Normalize the array[val] statistics.  */
+  if (array != NULL)
     {
-      /* Caught an arithmetic exception.  */
-      const char *msg;
+      unsigned long val;
 
-      /* End of protection against arithmetic exceptions.  */
-      uninstall_sigfpe_handler ();
-
-#if USE_SIGINFO
-      switch (sigfpe_code)
-#endif
-        {
-#if USE_SIGINFO
-# ifdef FPE_INTDIV
-        case FPE_INTDIV:
-          msg = _("plural expression can produce division by zero");
-          break;
-# endif
-# ifdef FPE_INTOVF
-        case FPE_INTOVF:
-          msg = _("plural expression can produce integer overflow");
-          break;
-# endif
-        default:
-#endif
-          msg = _("plural expression can produce arithmetic exceptions, possibly division by zero");
-        }
-
-      po_xerror (PO_SEVERITY_ERROR, header, NULL, 0, 0, false, msg);
-
-      free (array);
-
-      return 1;
+      for (val = 0; val < nplurals_value; val++)
+        array[val] = (array[val] == OFTEN ? 1 : 0);
     }
+
+  distribution->expr = plural_expr;
+  distribution->often = array;
+  distribution->often_length = (array != NULL ? nplurals_value : 0);
+  distribution->histogram = plural_expression_histogram;
+
+  return 0;
 #undef OFTEN
 }
 
@@ -288,7 +246,8 @@ static int
 check_plural (message_list_ty *mlp,
               int ignore_untranslated_messages,
               int ignore_fuzzy_messages,
-              struct plural_distribution *distributionp)
+              struct plural_distribution *distributionp,
+              xerror_handler_ty xeh)
 {
   int seen_errors = 0;
   const message_ty *has_plural;
@@ -368,16 +327,16 @@ check_plural (message_list_ty *mlp,
           if (help != NULL)
             {
               char *msg2ext = xasprintf ("%s\n%s", msg2, help);
-              po_xerror2 (PO_SEVERITY_ERROR,
-                          has_plural, NULL, 0, 0, false, msg1,
-                          header, NULL, 0, 0, true, msg2ext);
+              xeh->xerror2 (CAT_SEVERITY_ERROR,
+                            has_plural, NULL, 0, 0, false, msg1,
+                            header, NULL, 0, 0, true, msg2ext);
               free (msg2ext);
               free (help);
             }
           else
-            po_xerror2 (PO_SEVERITY_ERROR,
-                        has_plural, NULL, 0, 0, false, msg1,
-                        header, NULL, 0, 0, false, msg2);
+            xeh->xerror2 (CAT_SEVERITY_ERROR,
+                          has_plural, NULL, 0, 0, false, msg1,
+                          header, NULL, 0, 0, false, msg2);
 
           seen_errors++;
         }
@@ -392,16 +351,16 @@ check_plural (message_list_ty *mlp,
           if (help != NULL)
             {
               char *msg2ext = xasprintf ("%s\n%s", msg2, help);
-              po_xerror2 (PO_SEVERITY_ERROR,
-                          has_plural, NULL, 0, 0, false, msg1,
-                          header, NULL, 0, 0, true, msg2ext);
+              xeh->xerror2 (CAT_SEVERITY_ERROR,
+                            has_plural, NULL, 0, 0, false, msg1,
+                            header, NULL, 0, 0, true, msg2ext);
               free (msg2ext);
               free (help);
             }
           else
-            po_xerror2 (PO_SEVERITY_ERROR,
-                        has_plural, NULL, 0, 0, false, msg1,
-                        header, NULL, 0, 0, false, msg2);
+            xeh->xerror2 (CAT_SEVERITY_ERROR,
+                          has_plural, NULL, 0, 0, false, msg1,
+                          header, NULL, 0, 0, false, msg2);
 
           seen_errors++;
         }
@@ -428,13 +387,14 @@ check_plural (message_list_ty *mlp,
               if (help != NULL)
                 {
                   char *msgext = xasprintf ("%s\n%s", msg, help);
-                  po_xerror (PO_SEVERITY_ERROR, header, NULL, 0, 0, true,
-                             msgext);
+                  xeh->xerror (CAT_SEVERITY_ERROR, header, NULL, 0, 0, true,
+                               msgext);
                   free (msgext);
                   free (help);
                 }
               else
-                po_xerror (PO_SEVERITY_ERROR, header, NULL, 0, 0, false, msg);
+                xeh->xerror (CAT_SEVERITY_ERROR, header, NULL, 0, 0, false,
+                             msg);
 
               seen_errors++;
             }
@@ -450,13 +410,14 @@ check_plural (message_list_ty *mlp,
               if (help != NULL)
                 {
                   char *msgext = xasprintf ("%s\n%s", msg, help);
-                  po_xerror (PO_SEVERITY_ERROR, header, NULL, 0, 0, true,
-                             msgext);
+                  xeh->xerror (CAT_SEVERITY_ERROR, header, NULL, 0, 0, true,
+                               msgext);
                   free (msgext);
                   free (help);
                 }
               else
-                po_xerror (PO_SEVERITY_ERROR, header, NULL, 0, 0, false, msg);
+                xeh->xerror (CAT_SEVERITY_ERROR, header, NULL, 0, 0, false,
+                             msg);
 
               seen_errors++;
             }
@@ -466,7 +427,7 @@ check_plural (message_list_ty *mlp,
           if (!seen_errors)
             seen_errors =
               check_plural_eval (plural_expr, nplurals_value, header,
-                                 &distribution);
+                                 &distribution, xeh);
 
           /* Check the number of plurals of the translations.  */
           if (!seen_errors)
@@ -480,9 +441,9 @@ check_plural (message_list_ty *mlp,
                                          "but some messages have only %lu plural forms",
                                          min_nplurals),
                                min_nplurals);
-                  po_xerror2 (PO_SEVERITY_ERROR,
-                              header, NULL, 0, 0, false, msg1,
-                              min_pos, NULL, 0, 0, false, msg2);
+                  xeh->xerror2 (CAT_SEVERITY_ERROR,
+                                header, NULL, 0, 0, false, msg1,
+                                min_pos, NULL, 0, 0, false, msg2);
                   free (msg2);
                   free (msg1);
                   seen_errors++;
@@ -496,9 +457,9 @@ check_plural (message_list_ty *mlp,
                                          "but some messages have %lu plural forms",
                                          max_nplurals),
                                max_nplurals);
-                  po_xerror2 (PO_SEVERITY_ERROR,
-                              header, NULL, 0, 0, false, msg1,
-                              max_pos, NULL, 0, 0, false, msg2);
+                  xeh->xerror2 (CAT_SEVERITY_ERROR,
+                                header, NULL, 0, 0, false, msg1,
+                                max_pos, NULL, 0, 0, false, msg2);
                   free (msg2);
                   free (msg1);
                   seen_errors++;
@@ -515,8 +476,8 @@ check_plural (message_list_ty *mlp,
     {
       if (has_plural != NULL)
         {
-          po_xerror (PO_SEVERITY_ERROR, has_plural, NULL, 0, 0, false,
-                     _("message catalog has plural form translations, but lacks a header entry with \"Plural-Forms: nplurals=INTEGER; plural=EXPRESSION;\""));
+          xeh->xerror (CAT_SEVERITY_ERROR, has_plural, NULL, 0, 0, false,
+                       _("message catalog has plural form translations, but lacks a header entry with \"Plural-Forms: nplurals=INTEGER; plural=EXPRESSION;\""));
           seen_errors++;
         }
      no_plural:
@@ -545,27 +506,35 @@ check_plural (message_list_ty *mlp,
 
 
 /* Signal an error when checking format strings.  */
-static const message_ty *curr_mp;
-static lex_pos_ty curr_msgid_pos;
+struct formatstring_error_logger_locals
+{
+  xerror_handler_ty xeh;
+  const message_ty *curr_mp;
+  lex_pos_ty curr_msgid_pos;
+};
 static void
-formatstring_error_logger (const char *format, ...)
+formatstring_error_logger (void *data, const char *format, ...)
 #if defined __GNUC__ && ((__GNUC__ == 2 && __GNUC_MINOR__ >= 7) || __GNUC__ > 2)
-     __attribute__ ((__format__ (__printf__, 1, 2)))
+     __attribute__ ((__format__ (__printf__, 2, 3)))
 #endif
 ;
 static void
-formatstring_error_logger (const char *format, ...)
+formatstring_error_logger (void *data, const char *format, ...)
 {
+  struct formatstring_error_logger_locals *l =
+    (struct formatstring_error_logger_locals *) data;
   va_list args;
   char *msg;
 
   va_start (args, format);
   if (vasprintf (&msg, format, args) < 0)
-    error (EXIT_FAILURE, 0, _("memory exhausted"));
+    l->xeh->xerror (CAT_SEVERITY_FATAL_ERROR, NULL, NULL, 0, 0, false,
+                    _("memory exhausted"));
   va_end (args);
-  po_xerror (PO_SEVERITY_ERROR,
-             curr_mp, curr_msgid_pos.file_name, curr_msgid_pos.line_number,
-             (size_t)(-1), false, msg);
+  l->xeh->xerror (CAT_SEVERITY_ERROR,
+                  l->curr_mp,
+                  l->curr_msgid_pos.file_name, l->curr_msgid_pos.line_number,
+                  (size_t)(-1), false, msg);
   free (msg);
 }
 
@@ -587,7 +556,8 @@ check_pair (const message_ty *mp,
             int check_format_strings,
             const struct plural_distribution *distribution,
             int check_compatibility,
-            int check_accelerators, char accelerator_char)
+            int check_accelerators, char accelerator_char,
+            xerror_handler_ty xeh)
 {
   int seen_errors;
   int has_newline;
@@ -611,21 +581,21 @@ check_pair (const message_ty *mp,
 
           if (TEST_NEWLINE(msgid_plural) != has_newline)
             {
-              po_xerror (PO_SEVERITY_ERROR,
-                         mp, msgid_pos->file_name, msgid_pos->line_number,
-                         (size_t)(-1), false, _("\
-'msgid' and 'msgid_plural' entries do not both begin with '\\n'"));
+              xeh->xerror (CAT_SEVERITY_ERROR,
+                           mp, msgid_pos->file_name, msgid_pos->line_number,
+                           (size_t)(-1), false,
+                           _("'msgid' and 'msgid_plural' entries do not both begin with '\\n'"));
               seen_errors++;
             }
           for (p = msgstr, j = 0; p < msgstr + msgstr_len; p += strlen (p) + 1, j++)
             if (TEST_NEWLINE(p) != has_newline)
               {
                 char *msg =
-                  xasprintf (_("\
-'msgid' and 'msgstr[%u]' entries do not both begin with '\\n'"), j);
-                po_xerror (PO_SEVERITY_ERROR,
-                           mp, msgid_pos->file_name, msgid_pos->line_number,
-                           (size_t)(-1), false, msg);
+                  xasprintf (_("'msgid' and 'msgstr[%u]' entries do not both begin with '\\n'"),
+                             j);
+                xeh->xerror (CAT_SEVERITY_ERROR,
+                             mp, msgid_pos->file_name, msgid_pos->line_number,
+                             (size_t)(-1), false, msg);
                 free (msg);
                 seen_errors++;
               }
@@ -634,10 +604,10 @@ check_pair (const message_ty *mp,
         {
           if (TEST_NEWLINE(msgstr) != has_newline)
             {
-              po_xerror (PO_SEVERITY_ERROR,
-                         mp, msgid_pos->file_name, msgid_pos->line_number,
-                         (size_t)(-1), false, _("\
-'msgid' and 'msgstr' entries do not both begin with '\\n'"));
+              xeh->xerror (CAT_SEVERITY_ERROR,
+                           mp, msgid_pos->file_name, msgid_pos->line_number,
+                           (size_t)(-1), false,
+                           _("'msgid' and 'msgstr' entries do not both begin with '\\n'"));
               seen_errors++;
             }
         }
@@ -652,21 +622,21 @@ check_pair (const message_ty *mp,
 
           if (TEST_NEWLINE(msgid_plural) != has_newline)
             {
-              po_xerror (PO_SEVERITY_ERROR,
-                         mp, msgid_pos->file_name, msgid_pos->line_number,
-                         (size_t)(-1), false, _("\
-'msgid' and 'msgid_plural' entries do not both end with '\\n'"));
+              xeh->xerror (CAT_SEVERITY_ERROR,
+                           mp, msgid_pos->file_name, msgid_pos->line_number,
+                           (size_t)(-1), false,
+                           _("'msgid' and 'msgid_plural' entries do not both end with '\\n'"));
               seen_errors++;
             }
           for (p = msgstr, j = 0; p < msgstr + msgstr_len; p += strlen (p) + 1, j++)
             if (TEST_NEWLINE(p) != has_newline)
               {
                 char *msg =
-                  xasprintf (_("\
-'msgid' and 'msgstr[%u]' entries do not both end with '\\n'"), j);
-                po_xerror (PO_SEVERITY_ERROR,
-                           mp, msgid_pos->file_name, msgid_pos->line_number,
-                           (size_t)(-1), false, msg);
+                  xasprintf (_("'msgid' and 'msgstr[%u]' entries do not both end with '\\n'"),
+                             j);
+                xeh->xerror (CAT_SEVERITY_ERROR,
+                             mp, msgid_pos->file_name, msgid_pos->line_number,
+                             (size_t)(-1), false, msg);
                 free (msg);
                 seen_errors++;
               }
@@ -675,10 +645,10 @@ check_pair (const message_ty *mp,
         {
           if (TEST_NEWLINE(msgstr) != has_newline)
             {
-              po_xerror (PO_SEVERITY_ERROR,
-                         mp, msgid_pos->file_name, msgid_pos->line_number,
-                         (size_t)(-1), false, _("\
-'msgid' and 'msgstr' entries do not both end with '\\n'"));
+              xeh->xerror (CAT_SEVERITY_ERROR,
+                           mp, msgid_pos->file_name, msgid_pos->line_number,
+                           (size_t)(-1), false,
+                           _("'msgid' and 'msgstr' entries do not both end with '\\n'"));
               seen_errors++;
             }
         }
@@ -687,10 +657,10 @@ check_pair (const message_ty *mp,
 
   if (check_compatibility && msgid_plural != NULL)
     {
-      po_xerror (PO_SEVERITY_ERROR,
-                 mp, msgid_pos->file_name, msgid_pos->line_number,
-                 (size_t)(-1), false, _("\
-plural handling is a GNU gettext extension"));
+      xeh->xerror (CAT_SEVERITY_ERROR,
+                   mp, msgid_pos->file_name, msgid_pos->line_number,
+                   (size_t)(-1), false,
+                   _("plural handling is a GNU gettext extension"));
       seen_errors++;
     }
 
@@ -698,12 +668,14 @@ plural handling is a GNU gettext extension"));
     /* Test 3: Check whether both formats strings contain the same number
        of format specifications.  */
     {
-      curr_mp = mp;
-      curr_msgid_pos = *msgid_pos;
+      struct formatstring_error_logger_locals locals;
+      locals.xeh = xeh;
+      locals.curr_mp = mp;
+      locals.curr_msgid_pos = *msgid_pos;
       seen_errors +=
         check_msgid_msgstr_format (msgid, msgid_plural, msgstr, msgstr_len,
                                    is_format, mp->range, distribution,
-                                   formatstring_error_logger);
+                                   formatstring_error_logger, &locals);
     }
 
   if (check_accelerators && msgid_plural == NULL)
@@ -733,9 +705,9 @@ plural handling is a GNU gettext extension"));
               char *msg =
                 xasprintf (_("msgstr lacks the keyboard accelerator mark '%c'"),
                            accelerator_char);
-              po_xerror (PO_SEVERITY_ERROR,
-                         mp, msgid_pos->file_name, msgid_pos->line_number,
-                         (size_t)(-1), false, msg);
+              xeh->xerror (CAT_SEVERITY_ERROR,
+                           mp, msgid_pos->file_name, msgid_pos->line_number,
+                           (size_t)(-1), false, msg);
               free (msg);
               seen_errors++;
             }
@@ -744,9 +716,9 @@ plural handling is a GNU gettext extension"));
               char *msg =
                 xasprintf (_("msgstr has too many keyboard accelerator marks '%c'"),
                            accelerator_char);
-              po_xerror (PO_SEVERITY_ERROR,
-                         mp, msgid_pos->file_name, msgid_pos->line_number,
-                         (size_t)(-1), false, msg);
+              xeh->xerror (CAT_SEVERITY_ERROR,
+                           mp, msgid_pos->file_name, msgid_pos->line_number,
+                           (size_t)(-1), false, msg);
               free (msg);
               seen_errors++;
             }
@@ -759,7 +731,8 @@ plural handling is a GNU gettext extension"));
 
 /* Perform miscellaneous checks on a header entry.  */
 static int
-check_header_entry (const message_ty *mp, const char *msgstr_string)
+check_header_entry (const message_ty *mp, const char *msgstr_string,
+                    xerror_handler_ty xeh)
 {
   static const char *required_fields[] =
   {
@@ -789,10 +762,10 @@ check_header_entry (const message_ty *mp, const char *msgstr_string)
     {
 #if 0
       int severity =
-        (cnt < nrequiredfields ? PO_SEVERITY_ERROR : PO_SEVERITY_WARNING);
+        (cnt < nrequiredfields ? CAT_SEVERITY_ERROR : CAT_SEVERITY_WARNING);
 #else
       int severity =
-        PO_SEVERITY_WARNING;
+        CAT_SEVERITY_WARNING;
 #endif
       const char *field = required_fields[cnt];
       size_t len = strlen (field);
@@ -818,9 +791,9 @@ check_header_entry (const message_ty *mp, const char *msgstr_string)
                       char *msg =
                         xasprintf (_("header field '%s' still has the initial default value\n"),
                                    field);
-                      po_xerror (severity, mp, NULL, 0, 0, true, msg);
+                      xeh->xerror (severity, mp, NULL, 0, 0, true, msg);
                       free (msg);
-                      if (severity == PO_SEVERITY_ERROR)
+                      if (severity == CAT_SEVERITY_ERROR)
                         seen_errors++;
                     }
                 }
@@ -835,9 +808,9 @@ check_header_entry (const message_ty *mp, const char *msgstr_string)
           char *msg =
             xasprintf (_("header field '%s' missing in header\n"),
                        field);
-          po_xerror (severity, mp, NULL, 0, 0, true, msg);
+          xeh->xerror (severity, mp, NULL, 0, 0, true, msg);
           free (msg);
-          if (severity == PO_SEVERITY_ERROR)
+          if (severity == CAT_SEVERITY_ERROR)
             seen_errors++;
         }
     }
@@ -855,12 +828,13 @@ check_message (const message_ty *mp,
                const struct plural_distribution *distribution,
                int check_header,
                int check_compatibility,
-               int check_accelerators, char accelerator_char)
+               int check_accelerators, char accelerator_char,
+               xerror_handler_ty xeh)
 {
   int seen_errors = 0;
 
   if (check_header && is_header (mp))
-    seen_errors += check_header_entry (mp, mp->msgstr);
+    seen_errors += check_header_entry (mp, mp->msgstr, xeh);
 
   seen_errors += check_pair (mp,
                              mp->msgid, msgid_pos, mp->msgid_plural,
@@ -870,7 +844,8 @@ check_message (const message_ty *mp,
                              check_format_strings,
                              distribution,
                              check_compatibility,
-                             check_accelerators, accelerator_char);
+                             check_accelerators, accelerator_char,
+                             xeh);
   return seen_errors;
 }
 
@@ -885,7 +860,8 @@ check_message_list (message_list_ty *mlp,
                     int check_format_strings,
                     int check_header,
                     int check_compatibility,
-                    int check_accelerators, char accelerator_char)
+                    int check_accelerators, char accelerator_char,
+                    xerror_handler_ty xeh)
 {
   int seen_errors = 0;
   struct plural_distribution distribution;
@@ -898,7 +874,7 @@ check_message_list (message_list_ty *mlp,
 
   if (check_header)
     seen_errors += check_plural (mlp, ignore_untranslated_messages,
-                                 ignore_fuzzy_messages, &distribution);
+                                 ignore_fuzzy_messages, &distribution, xeh);
 
   for (j = 0; j < mlp->nitems; j++)
     {
@@ -912,208 +888,8 @@ check_message_list (message_list_ty *mlp,
                                       check_format_strings,
                                       &distribution,
                                       check_header, check_compatibility,
-                                      check_accelerators, accelerator_char);
-    }
-
-  return seen_errors;
-}
-
-
-static int
-syntax_check_ellipsis_unicode (const message_ty *mp, const char *msgid)
-{
-  const char *str = msgid;
-  const char *str_limit = str + strlen (msgid);
-  int seen_errors = 0;
-
-  while (str < str_limit)
-    {
-      const char *end, *cp;
-      ucs4_t ending_char;
-
-      end = sentence_end (str, &ending_char);
-
-      /* sentence_end doesn't treat '...' specially.  */
-      cp = end - (ending_char == '.' ? 2 : 3);
-      if (cp >= str && memcmp (cp, "...", 3) == 0)
-        {
-          po_xerror (PO_SEVERITY_ERROR, mp, NULL, 0, 0, false,
-                     _("ASCII ellipsis ('...') instead of Unicode"));
-          seen_errors++;
-        }
-
-      str = end + 1;
-    }
-
-  return seen_errors;
-}
-
-
-static int
-syntax_check_space_ellipsis (const message_ty *mp, const char *msgid)
-{
-  const char *str = msgid;
-  const char *str_limit = str + strlen (msgid);
-  int seen_errors = 0;
-
-  while (str < str_limit)
-    {
-      const char *end, *ellipsis = NULL;
-      ucs4_t ending_char;
-
-      end = sentence_end (str, &ending_char);
-
-      if (ending_char == 0x2026)
-        ellipsis = end;
-      else if (ending_char == '.')
-        {
-          /* sentence_end doesn't treat '...' specially.  */
-          const char *cp = end - 2;
-          if (cp >= str && memcmp (cp, "...", 3) == 0)
-            ellipsis = cp;
-        }
-      else
-        {
-          /* Look for a '...'.  */
-          const char *cp = end - 3;
-          if (cp >= str && memcmp (cp, "...", 3) == 0)
-            ellipsis = cp;
-          else
-            {
-              ucs4_t uc = 0xfffd;
-
-              /* Look for a U+2026.  */
-              for (cp = end - 1; cp >= str; cp--)
-                {
-                  u8_mbtouc (&uc, (const unsigned char *) cp, ellipsis - cp);
-                  if (uc != 0xfffd)
-                    break;
-                }
-
-              if (uc == 0x2026)
-                ellipsis = cp;
-            }
-        }
-
-      if (ellipsis)
-        {
-          const char *cp;
-          ucs4_t uc = 0xfffd;
-
-          /* Look at the character before ellipsis.  */
-          for (cp = ellipsis - 1; cp >= str; cp--)
-            {
-              u8_mbtouc (&uc, (const unsigned char *) cp, ellipsis - cp);
-              if (uc != 0xfffd)
-                break;
-            }
-
-          if (uc != 0xfffd && uc_is_space (uc))
-            {
-              po_xerror (PO_SEVERITY_ERROR, mp, NULL, 0, 0, false,
-                         _("\
-space before ellipsis found in user visible strings"));
-              seen_errors++;
-            }
-        }
-
-      str = end + 1;
-    }
-
-  return seen_errors;
-}
-
-
-struct callback_arg
-{
-  const message_ty *mp;
-  int seen_errors;
-};
-
-static void
-syntax_check_quote_unicode_callback (char quote, const char *quoted,
-                                     size_t quoted_length, void *data)
-{
-  struct callback_arg *arg = data;
-
-  switch (quote)
-    {
-    case '"':
-      po_xerror (PO_SEVERITY_ERROR, arg->mp, NULL, 0, 0, false,
-                 _("ASCII double quote used instead of Unicode"));
-      arg->seen_errors++;
-      break;
-
-    case '\'':
-      po_xerror (PO_SEVERITY_ERROR, arg->mp, NULL, 0, 0, false,
-                 _("ASCII single quote used instead of Unicode"));
-      arg->seen_errors++;
-      break;
-
-    default:
-      break;
-    }
-}
-
-static int
-syntax_check_quote_unicode (const message_ty *mp, const char *msgid)
-{
-  struct callback_arg arg;
-
-  arg.mp = mp;
-  arg.seen_errors = 0;
-
-  scan_quoted (msgid, strlen (msgid),
-               syntax_check_quote_unicode_callback, &arg);
-
-  return arg.seen_errors;
-}
-
-
-typedef int (* syntax_check_function) (const message_ty *mp, const char *msgid);
-static const syntax_check_function sc_funcs[NSYNTAXCHECKS] =
-{
-  syntax_check_ellipsis_unicode,
-  syntax_check_space_ellipsis,
-  syntax_check_quote_unicode
-};
-
-/* Perform all syntax checks on a non-obsolete message.
-   Return the number of errors that were seen.  */
-static int
-syntax_check_message (const message_ty *mp)
-{
-  int seen_errors = 0;
-  int i;
-
-  for (i = 0; i < NSYNTAXCHECKS; i++)
-    {
-      if (mp->do_syntax_check[i] == yes)
-        {
-          seen_errors += sc_funcs[i] (mp, mp->msgid);
-          if (mp->msgid_plural)
-            seen_errors += sc_funcs[i] (mp, mp->msgid_plural);
-        }
-    }
-
-  return seen_errors;
-}
-
-
-/* Perform all syntax checks on a message list.
-   Return the number of errors that were seen.  */
-int
-syntax_check_message_list (message_list_ty *mlp)
-{
-  int seen_errors = 0;
-  size_t j;
-
-  for (j = 0; j < mlp->nitems; j++)
-    {
-      message_ty *mp = mlp->item[j];
-
-      if (!is_header (mp))
-        seen_errors += syntax_check_message (mp);
+                                      check_accelerators, accelerator_char,
+                                      xeh);
     }
 
   return seen_errors;
