@@ -1,5 +1,5 @@
 /* Get the contents of an URL.
-   Copyright (C) 2001-2003, 2005-2010, 2012, 2017-2023 Free Software Foundation, Inc.
+   Copyright (C) 2001-2024 Free Software Foundation, Inc.
    Written by Bruno Haible <haible@clisp.cons.org>, 2001.
 
    This program is free software: you can redistribute it and/or modify
@@ -29,9 +29,9 @@
 #include <locale.h>
 #include <unistd.h>
 
+#include <error.h>
 #include "noreturn.h"
 #include "closeout.h"
-#include "error.h"
 #include "error-progname.h"
 #include "progname.h"
 #include "relocatable.h"
@@ -101,6 +101,7 @@ main (int argc, char *argv[])
 
   /* Set the text message domain.  */
   bindtextdomain (PACKAGE, relocate (LOCALEDIR));
+  bindtextdomain ("gnulib", relocate (GNULIB_LOCALEDIR));
   textdomain (PACKAGE);
 
   /* Ensure that write errors on stdout are detected.  */
@@ -141,7 +142,7 @@ License GPLv3+: GNU GPL version 3 or later <%s>\n\
 This is free software: you are free to change and redistribute it.\n\
 There is NO WARRANTY, to the extent permitted by law.\n\
 "),
-              "2001-2023", "https://gnu.org/licenses/gpl.html");
+              "2001-2024", "https://gnu.org/licenses/gpl.html");
       printf (_("Written by %s.\n"), proper_name ("Bruno Haible"));
       exit (EXIT_SUCCESS);
     }
@@ -251,7 +252,7 @@ execute_it (const char *progname,
   (void) private_data;
 
   java_exitcode =
-    execute (progname, prog_path, prog_argv, NULL,
+    execute (progname, prog_path, prog_argv, NULL, NULL,
              true, true, false, false, true, false, NULL);
   /* Exit code 0 means success, 2 means timed out.  */
   return !(java_exitcode == 0 || java_exitcode == 2);
@@ -319,7 +320,7 @@ fetch (const char *url, const char *file)
         argv[0] = "wget";
         argv[1] = "--version";
         argv[2] = NULL;
-        exitstatus = execute ("wget", "wget", argv, NULL,
+        exitstatus = execute ("wget", "wget", argv, NULL, NULL,
                               false, false, true, true, true, false, NULL);
         wget_present = (exitstatus == 0);
         wget_tested = true;
@@ -337,7 +338,7 @@ fetch (const char *url, const char *file)
         argv[6] = "--user-agent"; argv[7] = "urlget";
         argv[8] = url;
         argv[9] = NULL;
-        exitstatus = execute ("wget", "wget", argv, NULL,
+        exitstatus = execute ("wget", "wget", argv, NULL, NULL,
                               true, false, false, false, true, false, NULL);
         if (exitstatus != 127)
           {
@@ -364,7 +365,7 @@ fetch (const char *url, const char *file)
         argv[0] = "lynx";
         argv[1] = "--version";
         argv[2] = NULL;
-        exitstatus = execute ("lynx", "lynx", argv, NULL,
+        exitstatus = execute ("lynx", "lynx", argv, NULL, NULL,
                               false, false, true, true, true, false, NULL);
         lynx_present = (exitstatus == 0);
         lynx_tested = true;
@@ -380,7 +381,7 @@ fetch (const char *url, const char *file)
         argv[2] = "-source";
         argv[3] = url;
         argv[4] = NULL;
-        exitstatus = execute ("lynx", "lynx", argv, NULL,
+        exitstatus = execute ("lynx", "lynx", argv, NULL, NULL,
                               true, false, false, false, true, false, NULL);
         if (exitstatus != 127)
           {
@@ -407,7 +408,7 @@ fetch (const char *url, const char *file)
         argv[0] = "curl";
         argv[1] = "--version";
         argv[2] = NULL;
-        exitstatus = execute ("curl", "curl", argv, NULL,
+        exitstatus = execute ("curl", "curl", argv, NULL, NULL,
                               false, false, true, true, true, false, NULL);
         curl_present = (exitstatus == 0 || exitstatus == 2);
         curl_tested = true;
@@ -423,7 +424,7 @@ fetch (const char *url, const char *file)
         argv[2] = "--user-agent"; argv[3] = "urlget";
         argv[4] = url;
         argv[5] = NULL;
-        exitstatus = execute ("curl", "curl", argv, NULL,
+        exitstatus = execute ("curl", "curl", argv, NULL, NULL,
                               true, false, false, false, true, false, NULL);
         if (exitstatus != 127)
           {
