@@ -1,5 +1,5 @@
 /* xgettext JavaScript backend.
-   Copyright (C) 2002-2003, 2005-2009, 2013-2014, 2018-2020 Free Software Foundation, Inc.
+   Copyright (C) 2002-2024 Free Software Foundation, Inc.
 
    This file was written by Andreas Stricker <andy@knitter.ch>, 2010
    It's based on x-python from Bruno Haible.
@@ -31,8 +31,10 @@
 #include <stdlib.h>
 #include <string.h>
 
+#include <error.h>
 #include "attribute.h"
 #include "message.h"
+#include "str-list.h"
 #include "rc-str-list.h"
 #include "xgettext.h"
 #include "xg-pos.h"
@@ -42,12 +44,13 @@
 #include "xg-arglist-callshape.h"
 #include "xg-arglist-parser.h"
 #include "xg-message.h"
-#include "error.h"
-#include "error-progname.h"
+#include "if-error.h"
+#include "xstrerror.h"
 #include "progname.h"
 #include "xerror.h"
 #include "xvasprintf.h"
 #include "xalloc.h"
+#include "string-buffer.h"
 #include "c-strstr.h"
 #include "c-ctype.h"
 #include "po-charset.h"
@@ -56,9 +59,6 @@
 
 #define _(s) gettext(s)
 
-#undef max /* clean up after MSVC's <stdlib.h> */
-#define max(a,b) ((a) > (b) ? (a) : (b))
-
 #define SIZEOF(a) (sizeof(a) / sizeof(a[0]))
 
 /* The JavaScript aka ECMA-Script syntax is defined in ECMA-262
@@ -151,6 +151,105 @@ init_flag_table_javascript ()
 }
 
 
+/* ======================== Tag set customization.  ======================== */
+
+/* Tagged template literals are described in
+   <https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Template_literals>.
+
+   A tagged template literal looks like this in the source code:
+     TAG`part0 ${expression 1} part1 ${expression 2} ... ${expression N} partN`
+
+   A tag, immediately before a template literal, denotes a function that takes
+   as arguments:
+     - A list of the N+1 parts of the template literal,
+     - The N values of the N expressions between these parts.
+
+   In our use case, the tag function is supposed to
+     1. Convert the N+1 parts to a format string.
+     2. Look up the translation of this format string. (It is supposed to
+        accept the same number of arguments.)
+     3. Call a formatting facility that substitutes the N values into the
+        translated format string.
+   The type of format string is not fixed.  */
+
+/* Type of a C function that implements step 1 of what a tag function does.
+   It takes a non-empty string_list_ty as argument and returns a freshly
+   allocated string.  */
+typedef char * (*tag_step1_fn) (string_list_ty *parts);
+
+/* Tag step 1 function that produces a format string with placeholders
+   {0}, {1}, {2}, etc.  */
+static char *
+gnome_step1 (string_list_ty *parts)
+{
+  size_t n = parts->nitems - 1;
+  string_list_ty pieces;
+  unsigned long i;
+
+  pieces.nitems = 2 * n + 1;
+  pieces.nitems_max = pieces.nitems;
+  pieces.item = XNMALLOC (pieces.nitems, const char *);
+  for (i = 0; i <= n; i++)
+    {
+      pieces.item[2 * i] = parts->item[i];
+      if (i < n)
+        pieces.item[2 * i + 1] = xasprintf ("{%lu}", i);
+    }
+
+  char *result = string_list_concat (&pieces);
+
+  for (i = 0; i < n; i++)
+    free ((char *) pieces.item[2 * i + 1]);
+
+  return result;
+}
+
+/* Returns the tag step 1 function for a given format, or NULL if that format
+   is unknown.  */
+static tag_step1_fn
+get_tag_step1_fn (const char *format)
+{
+  if (strcmp (format, "javascript-gnome-format") == 0)
+    return gnome_step1;
+  /* ... More formats can be added here ...  */
+  return NULL;
+}
+
+/* Information associated with a tag.  */
+struct tag_definition
+{
+  const char *format;
+  tag_step1_fn step1_fn;
+};
+
+/* Mapping tag -> format.  */
+static hash_table tags;
+
+void
+x_javascript_tag (const char *name)
+{
+  const char *colon = strchr (name, ':');
+  if (colon != NULL)
+    {
+      const char *format = colon + 1;
+      tag_step1_fn step1_fn = get_tag_step1_fn (format);
+      if (step1_fn != NULL)
+        {
+          /* Heap-allocate a 'struct tag_definition'  */
+          struct tag_definition *def = XMALLOC (struct tag_definition);
+          def->format = xstrdup (format);
+          def->step1_fn = step1_fn;
+
+          if (tags.table == NULL)
+            hash_init (&tags, 10);
+
+          /* Insert it in the TAGS table.  */
+          hash_set_value (&tags, name, colon - name, def);
+        }
+    }
+}
+
+
 /* ======================== Reading of characters.  ======================== */
 
 /* The input file stream.  */
@@ -215,8 +314,8 @@ phase1_ungetc (int c)
 
 static lexical_context_ty lexical_context;
 
-/* Maximum used, length of "<![CDATA[" tag minus one.  */
-static int phase2_pushback[8];
+/* Maximum used, length of "<![CDATA[" tag.  */
+static int phase2_pushback[9];
 static int phase2_pushback_length;
 
 /* Read the next Unicode UCS-4 character from the input file.  */
@@ -251,11 +350,14 @@ phase2_getc ()
          interactive behaviour when fp is connected to an interactive tty.  */
       unsigned char buf[MAX_PHASE1_PUSHBACK];
       size_t bufcount;
-      int c = phase1_getc ();
-      if (c == EOF)
-        return UEOF;
-      buf[0] = (unsigned char) c;
-      bufcount = 1;
+
+      {
+        int c = phase1_getc ();
+        if (c == EOF)
+          return UEOF;
+        buf[0] = (unsigned char) c;
+        bufcount = 1;
+      }
 
       for (;;)
         {
@@ -328,8 +430,9 @@ Please specify the correct source encoding through --from-code\n"),
                   buf[bufcount++] = (unsigned char) c;
                 }
               else
-                error (EXIT_FAILURE, errno, _("%s:%d: iconv failure"),
-                       real_file_name, line_number);
+                if_error (IF_SEVERITY_FATAL_ERROR,
+                          real_file_name, line_number, (size_t)(-1), false,
+                          "%s", xstrerror (_("iconv failure"), errno));
             }
           else
             {
@@ -441,7 +544,7 @@ Please specify the source encoding through --from-code\n"),
     }
 }
 
-/* Supports max (9, UNINAME_MAX + 3) pushback characters.  */
+/* Supports 9 pushback characters.  */
 static void
 phase2_ungetc (int c)
 {
@@ -696,9 +799,15 @@ typedef struct token_ty token_ty;
 struct token_ty
 {
   token_type_ty type;
-  char *string;                  /* for token_type_symbol, token_type_keyword */
-  mixed_string_ty *mixed_string;        /* for token_type_string, token_type_template */
-  refcounted_string_list_ty *comment;   /* for token_type_string, token_type_template */
+  char *template_tag;                 /* for token_type_template, token_type_ltemplate,
+                                         token_type_rtemplate */
+  char *string;                       /* for token_type_symbol, token_type_keyword */
+  mixed_string_ty *mixed_string;      /* for token_type_string, token_type_template,
+                                         token_type_ltemplate, token_type_mtemplate,
+                                         token_type_rtemplate */
+  string_list_ty *template_parts;     /* for token_type_rtemplate */
+  refcounted_string_list_ty *comment; /* for token_type_string, token_type_template,
+                                         token_type_ltemplate, token_type_rtemplate */
   int line_number;
 };
 
@@ -707,13 +816,23 @@ struct token_ty
 static inline void
 free_token (token_ty *tp)
 {
+  if (tp->type == token_type_template || tp->type == token_type_ltemplate
+      || tp->type == token_type_rtemplate)
+    free (tp->template_tag);
   if (tp->type == token_type_symbol || tp->type == token_type_keyword)
     free (tp->string);
-  if (tp->type == token_type_string || tp->type == token_type_template)
-    {
-      mixed_string_free (tp->mixed_string);
-      drop_reference (tp->comment);
-    }
+  if (tp->type == token_type_string || tp->type == token_type_template
+      /* For these types, tp->mixed_string is already freed earlier, when we
+         build up the level's template_parts.  */
+      #if 0
+      || tp->type == token_type_ltemplate || tp->type == token_type_mtemplate
+      || tp->type == token_type_rtemplate
+      #endif
+     )
+    mixed_string_free (tp->mixed_string);
+  if (tp->type == token_type_string || tp->type == token_type_template
+      || tp->type == token_type_ltemplate || tp->type == token_type_rtemplate)
+    drop_reference (tp->comment);
 }
 
 
@@ -762,10 +881,9 @@ phase7_getuc (int quote_char)
           else
             {
               phase2_ungetc (c);
-              error_with_progname = false;
-              error (0, 0, _("%s:%d: warning: unterminated string"),
-                     logical_file_name, line_number);
-              error_with_progname = true;
+              if_error (IF_SEVERITY_WARNING,
+                        logical_file_name, line_number, (size_t)(-1), false,
+                        _("unterminated string"));
               return P7_STRING_END;
             }
         }
@@ -976,11 +1094,9 @@ phase5_scan_regexp (void)
         }
       if (c == UEOF)
         {
-          error_with_progname = false;
-          error (0, 0,
-                 _("%s:%d: warning: RegExp literal terminated too early"),
-                 logical_file_name, line_number);
-          error_with_progname = true;
+          if_error (IF_SEVERITY_WARNING,
+                    logical_file_name, line_number, (size_t)(-1), false,
+                    _("RegExp literal terminated too early"));
           return;
         }
     }
@@ -991,47 +1107,79 @@ phase5_scan_regexp (void)
     phase2_ungetc (c);
 }
 
-/* Number of open template literals `...${  */
-static int template_literal_depth;
-
-/* Number of open '{' tokens, at each template literal level.
-   The "current" element is brace_depths[template_literal_depth].  */
-static int *brace_depths;
-/* Number of allocated elements in brace_depths.  */
-static size_t brace_depths_alloc;
-
-/* Adds a new brace_depths level after template_literal_depth was
-   incremented.  */
-static void
-new_brace_depth_level (void)
+/* Various syntactic constructs can be nested:
+     - braces in expressions       {
+     - template literals           `...${
+     - XML elements                <tag>
+     - embedded JavaScript in XML  {
+   For a well-formed program:
+     - expressions must have balanced braces;
+     - template literals must be closed before the embedded JavaScript is closed;
+     - the embedded JavaScript must be closed before the XML element is closed;
+     - and so on.
+   Therefore we can represent these nested syntactic constructs with a stack;
+   each element is a new level.  */
+enum level_ty
 {
-  if (template_literal_depth == brace_depths_alloc)
+  level_brace              = 1,
+  level_template_literal   = 2,
+  level_xml_element        = 3,
+  level_embedded_js_in_xml = 4
+};
+struct level_info
+{
+  enum level_ty type;
+  char *template_tag;                          /* for level_template_literal */
+  string_list_ty *template_parts;              /* for level_template_literal */
+  refcounted_string_list_ty *template_comment; /* for level_template_literal */
+};
+/* The level stack.  */
+static struct level_info *levels /* = NULL */;
+/* Number of allocated elements in levels.  */
+static size_t levels_alloc /* = 0 */;
+/* Number of currently used elements in levels.  */
+static size_t level;
+
+/* Adds a new level.  */
+static void
+new_level (enum level_ty l)
+{
+  if (level == levels_alloc)
     {
-      brace_depths_alloc = 2 * brace_depths_alloc + 1;
-      /* Now template_literal_depth < brace_depths_alloc.  */
-      brace_depths =
-        (int *) xrealloc (brace_depths, brace_depths_alloc * sizeof (int));
+      levels_alloc = 2 * levels_alloc + 1;
+      /* Now level < levels_alloc.  */
+      levels =
+        (struct level_info *)
+        xrealloc (levels, levels_alloc * sizeof (struct level_info));
     }
-  brace_depths[template_literal_depth] = 0;
+  levels[level].type = l;
+  level++;
 }
 
-/* Number of open XML elements.  */
-static int xml_element_depth;
-static bool inside_embedded_js_in_xml;
+/* Returns the current level's type,
+   as one of the level_* enum items, or 0 if the level stack is empty.  */
+#define level_type() \
+  (level > 0 ? levels[level - 1].type : 0)
 
-static bool
+/* Parses some XML markup.
+   Returns 0 for an XML comment,
+           1 for a CDATA,
+           2 for an XML Processing Instruction,
+   or -1 when none of them was recognized.  */
+static int
 phase5_scan_xml_markup (token_ty *tp)
 {
   struct
-  {
-    const char *start;
-    const char *end;
-  } markers[] =
-      {
-        { "!--", "--" },
-        { "![CDATA[", "]]" },
-        { "?", "?" }
-      };
+    {
+      const char *start;
+      const char *end;
+    }
+  markers[] =
+    {
+      { "!--", "--" },
+      { "![CDATA[", "]]" },
+      { "?", "?" }
+    };
   int i;
 
   for (i = 0; i < SIZEOF (markers); i++)
@@ -1099,27 +1247,22 @@ phase5_scan_xml_markup (token_ty *tp)
                   goto eof;
                 if (c != '>')
                   {
-                    error_with_progname = false;
-                    error (0, 0,
-                           _("%s:%d: warning: %s is not allowed"),
-                           logical_file_name, line_number,
-                           end);
-                    error_with_progname = true;
-                    return false;
+                    if_error (IF_SEVERITY_WARNING,
+                              logical_file_name, line_number, (size_t)(-1), false,
+                              _("%s is not allowed"), end);
+                    return -1;
                   }
-                return true;
+                return i;
               }
           }
     }
-  return false;
+  return -1;
 
  eof:
-  error_with_progname = false;
-  error (0, 0,
-         _("%s:%d: warning: unterminated XML markup"),
-         logical_file_name, line_number);
-  error_with_progname = true;
-  return false;
+  if_error (IF_SEVERITY_WARNING,
+            logical_file_name, line_number, (size_t)(-1), false,
+            _("unterminated XML markup"));
+  return -1;
 }
 
 static void
@@ -1187,19 +1330,12 @@ phase5_get (token_ty *tp)
         case '5': case '6': case '7': case '8': case '9':
           /* Symbol, or part of a number.  */
           {
-            static char *buffer;
-            static int bufmax;
-            int bufpos;
+            struct string_buffer buffer;
 
-            bufpos = 0;
+            sb_init (&buffer);
             for (;;)
               {
-                if (bufpos >= bufmax)
-                  {
-                    bufmax = 2 * bufmax + 10;
-                    buffer = xrealloc (buffer, bufmax);
-                  }
-                buffer[bufpos++] = c;
+                sb_xappend1 (&buffer, c);
                 c = phase3_getc ();
                 switch (c)
                   {
@@ -1223,15 +1359,9 @@ phase5_get (token_ty *tp)
                   }
                 break;
               }
-            if (bufpos >= bufmax)
-              {
-                bufmax = 2 * bufmax + 10;
-                buffer = xrealloc (buffer, bufmax);
-              }
-            buffer[bufpos] = '\0';
-            tp->string = xstrdup (buffer);
-            if (strcmp (buffer, "return") == 0
-                || strcmp (buffer, "else") == 0)
+            tp->string = sb_xdupfree_c (&buffer);
+            if (strcmp (tp->string, "return") == 0
+                || strcmp (tp->string, "else") == 0)
               tp->type = last_token_type = token_type_keyword;
             else
               tp->type = last_token_type = token_type_symbol;
@@ -1294,6 +1424,7 @@ phase5_get (token_ty *tp)
 
                 if (uc == P7_EOF || uc == P7_STRING_END)
                   {
+                    tp->template_tag = NULL;
                     tp->mixed_string = mixed_string_buffer_result (&msb);
                     tp->comment = add_reference (savable_comment);
                     tp->type = last_token_type = token_type_template;
@@ -1302,10 +1433,14 @@ phase5_get (token_ty *tp)
 
                 if (uc == P7_TEMPLATE_START_OF_EXPRESSION)
                   {
-                    mixed_string_buffer_destroy (&msb);
+                    tp->template_tag = NULL;
+                    tp->mixed_string = mixed_string_buffer_result (&msb);
+                    tp->comment = add_reference (savable_comment);
                     tp->type = last_token_type = token_type_ltemplate;
-                    template_literal_depth++;
-                    new_brace_depth_level ();
+                    new_level (level_template_literal);
+                    levels[level - 1].template_tag = NULL;
+                    levels[level - 1].template_parts = NULL;
+                    levels[level - 1].template_comment = NULL;
                     break;
                   }
 
@@ -1346,16 +1481,25 @@ phase5_get (token_ty *tp)
                - XMLMarkup and XMLElement are not allowed after an expression,
                - embedded JavaScript expressions in XML do not recurse.
              */
-            if (xml_element_depth > 0
-                || (!inside_embedded_js_in_xml
+            if (level_type () == level_xml_element
+                || (level_type () != level_embedded_js_in_xml
                     && ! is_after_expression ()))
               {
-                /* Comments, PI, or CDATA.  */
-                if (phase5_scan_xml_markup (tp))
-                  /* BUG: *tp is not filled in here!  */
-                  return;
-                c = phase2_getc ();
+                /* Recognize XML markup: XML comment, CDATA, Processing
+                   Instruction.  */
+                int xml_markup_type = phase5_scan_xml_markup (tp);
+                if (xml_markup_type >= 0)
+                  {
+                    /* Ignore them all, since they are not part of JSX.
+                       But warn about CDATA.  */
+                    if (xml_markup_type == 1)
+                      if_error (IF_SEVERITY_WARNING,
+                                logical_file_name, line_number, (size_t)(-1), false,
+                                _("ignoring CDATA section"));
+                    continue;
+                  }
 
+                c = phase2_getc ();
                 if (c == '/')
                   {
                     /* Closing tag.  */
@@ -1366,7 +1510,7 @@ phase5_get (token_ty *tp)
                     /* Opening element.  */
                     phase2_ungetc (c);
                     lexical_context = lc_xml_open_tag;
-                    xml_element_depth++;
+                    new_level (level_xml_element);
                   }
                 tp->type = last_token_type = token_type_xml_tag;
               }
@@ -1376,7 +1520,7 @@ phase5_get (token_ty *tp)
           return;
 
         case '>':
-          if (xml_element_depth > 0 && !inside_embedded_js_in_xml)
+          if (level_type () == level_xml_element)
             {
               switch (lexical_context)
                 {
@@ -1386,7 +1530,8 @@ phase5_get (token_ty *tp)
                   return;
 
                 case lc_xml_close_tag:
-                  if (--xml_element_depth > 0)
+                  level--;
+                  if (memchr (levels, level_xml_element, level) != NULL)
                     lexical_context = lc_xml_content;
                   else
                     lexical_context = lc_outside;
@@ -1401,7 +1546,7 @@ phase5_get (token_ty *tp)
           return;
 
         case '/':
-          if (xml_element_depth > 0 && !inside_embedded_js_in_xml)
+          if (level_type () == level_xml_element)
             {
               /* If it appears in an opening tag of an XML element, it's
                  part of '/>'.  */
@@ -1410,7 +1555,8 @@ phase5_get (token_ty *tp)
                   c = phase2_getc ();
                   if (c == '>')
                     {
-                      if (--xml_element_depth > 0)
+                      level--;
+                      if (memchr (levels, level_xml_element, level) != NULL)
                         lexical_context = lc_xml_content;
                       else
                         lexical_context = lc_outside;
@@ -1435,38 +1581,66 @@ phase5_get (token_ty *tp)
           return;
 
         case '{':
-          if (xml_element_depth > 0 && !inside_embedded_js_in_xml)
-            inside_embedded_js_in_xml = true;
+          if (level_type () == level_xml_element)
+            new_level (level_embedded_js_in_xml);
           else
-            brace_depths[template_literal_depth]++;
+            new_level (level_brace);
           tp->type = last_token_type = token_type_lbrace;
           return;
 
         case '}':
-          if (xml_element_depth > 0 && inside_embedded_js_in_xml)
-            inside_embedded_js_in_xml = false;
-          else if (brace_depths[template_literal_depth] > 0)
-            brace_depths[template_literal_depth]--;
-          else if (template_literal_depth > 0)
+          if (level_type () == level_embedded_js_in_xml)
+            level--;
+          else if (level_type () == level_brace)
+            level--;
+          else if (level_type () == level_template_literal)
             {
               /* Middle or right part of template literal.  */
+              struct mixed_string_buffer msb;
+
+              lexical_context = lc_string;
+              /* Start accumulating the string.  */
+              mixed_string_buffer_init (&msb, lexical_context,
+                                        logical_file_name, line_number);
               for (;;)
                 {
                   int uc = phase7_getuc ('`');
 
+                  /* Keep line_number in sync.  */
+                  msb.line_number = line_number;
+
                   if (uc == P7_EOF || uc == P7_STRING_END)
                     {
+                      tp->mixed_string = mixed_string_buffer_result (&msb);
                       tp->type = last_token_type = token_type_rtemplate;
-                      template_literal_depth--;
+                      string_list_append_move (levels[level - 1].template_parts,
+                                               mixed_string_contents_free1 (tp->mixed_string));
+                      /* Move info from the current level to the token.  */
+                      tp->template_tag = levels[level - 1].template_tag;
+                      tp->template_parts = levels[level - 1].template_parts;
+                      tp->comment = levels[level - 1].template_comment;
+                      level--;
                       break;
                     }
 
                   if (uc == P7_TEMPLATE_START_OF_EXPRESSION)
                     {
+                      tp->mixed_string = mixed_string_buffer_result (&msb);
                       tp->type = last_token_type = token_type_mtemplate;
                       break;
                     }
+
+                  if (IS_UNICODE (uc))
+                    {
+                      assert (UNICODE_VALUE (uc) >= 0
+                              && UNICODE_VALUE (uc) < 0x110000);
+                      mixed_string_buffer_append_unicode (&msb,
+                                                          UNICODE_VALUE (uc));
+                    }
+                  else
+                    mixed_string_buffer_append_char (&msb, uc);
                 }
+              lexical_context = lc_outside;
               return;
             }
           tp->type = last_token_type = token_type_rbrace;
@@ -1522,6 +1696,7 @@ static void
 x_javascript_lex (token_ty *tp)
 {
   phase5_get (tp);
+
   if (tp->type == token_type_string || tp->type == token_type_template)
     {
       mixed_string_ty *sum = tp->mixed_string;
@@ -1557,19 +1732,40 @@ x_javascript_lex (token_ty *tp)
       token_ty token2;
 
       phase5_get (&token2);
-      if (token2.type == token_type_template)
+      if (token2.type == token_type_template
+          || token2.type == token_type_ltemplate)
         {
-          /* The value of
-               tag `abc`
-             is the value of the function call
-               tag (["abc"])
-             We don't know anything about this value.  Therefore, don't
-             let the extractor see this template literal.  */
-          free_token (&token2);
+          /* Merge *tp and token2:
+             tag `abc`    becomes    tag`abc`
+             tag `abc${   becomes    tag`abc${
+           */
+          tp->type = token2.type;
+          tp->template_tag = tp->string;
+          tp->mixed_string = token2.mixed_string;
+          tp->comment = token2.comment;
+          tp->line_number = token2.line_number;
         }
       else
         phase5_unget (&token2);
     }
+
+  /* Move info from the token into the current level.  */
+  if (tp->type == token_type_ltemplate
+      || tp->type == token_type_mtemplate)
+    {
+      if (!(level_type () == level_template_literal))
+        abort ();
+      if (tp->type == token_type_ltemplate)
+        {
+          levels[level - 1].template_tag = tp->template_tag;
+          tp->template_tag = NULL;
+          levels[level - 1].template_parts = string_list_alloc ();
+          levels[level - 1].template_comment = tp->comment;
+          tp->comment = NULL;
+        }
+      string_list_append_move (levels[level - 1].template_parts,
+                               mixed_string_contents_free1 (tp->mixed_string));
+    }
 }
 
 
@@ -1580,6 +1776,16 @@ x_javascript_lex (token_ty *tp)
 static flag_context_list_table_ty *flag_context_list_table;
 
 
+/* Maximum supported nesting depth.  */
+#define MAX_NESTING_DEPTH 1000
+
+/* Current nesting depths.  */
+static int paren_nesting_depth;
+static int bracket_nesting_depth;
+static int brace_nesting_depth;
+static int xml_element_nesting_depth;
+
+
 /* The file is broken into tokens.  Scan the token stream, looking for
    a keyword, followed by a left paren, followed by a string.  When we
    see this sequence, we have something to remember.  We assume we are
@@ -1603,7 +1809,7 @@ static flag_context_list_table_ty *flag_context_list_table;
 static bool
 extract_balanced (message_list_ty *mlp,
                   token_type_ty delim,
-                  flag_context_ty outer_context,
+                  flag_region_ty *outer_region,
                   flag_context_list_iterator_ty context_iter,
                   struct arglist_parser *argparser)
 {
@@ -1616,9 +1822,9 @@ extract_balanced (message_list_ty *mlp,
   /* Context iterator that will be used if the next token is a '('.  */
   flag_context_list_iterator_ty next_context_iter =
     passthrough_context_list_iterator;
-  /* Current context.  */
-  flag_context_ty inner_context =
-    inherited_context (outer_context,
+  /* Current region.  */
+  flag_region_ty *inner_region =
+    inheriting_region (outer_region,
                        flag_context_list_iterator_advance (&context_iter));
 
   /* Start state is 0.  */
@@ -1654,14 +1860,20 @@ extract_balanced (message_list_ty *mlp,
           continue;
 
         case token_type_lparen:
+          if (++paren_nesting_depth > MAX_NESTING_DEPTH)
+            if_error (IF_SEVERITY_FATAL_ERROR,
+                      logical_file_name, line_number, (size_t)(-1), false,
+                      _("too many open parentheses"));
           if (extract_balanced (mlp, token_type_rparen,
-                                inner_context, next_context_iter,
+                                inner_region, next_context_iter,
                                 arglist_parser_alloc (mlp,
                                                       state ? next_shapes : NULL)))
             {
               arglist_parser_done (argparser, arg);
+              unref_region (inner_region);
               return true;
             }
+          paren_nesting_depth--;
           next_context_iter = null_context_list_iterator;
           state = 0;
           continue;
@@ -1670,6 +1882,7 @@ extract_balanced (message_list_ty *mlp,
           if (delim == token_type_rparen || delim == token_type_eof)
             {
               arglist_parser_done (argparser, arg);
+              unref_region (inner_region);
               return false;
             }
           next_context_iter = null_context_list_iterator;
@@ -1678,8 +1891,9 @@ extract_balanced (message_list_ty *mlp,
 
         case token_type_comma:
           arg++;
-          inner_context =
-            inherited_context (outer_context,
+          unref_region (inner_region);
+          inner_region =
+            inheriting_region (outer_region,
                                flag_context_list_iterator_advance (
                                  &context_iter));
           next_context_iter = passthrough_context_list_iterator;
@@ -1687,13 +1901,20 @@ extract_balanced (message_list_ty *mlp,
           continue;
 
         case token_type_lbracket:
+          if (++bracket_nesting_depth > MAX_NESTING_DEPTH)
+            if_error (IF_SEVERITY_FATAL_ERROR,
+                      logical_file_name, line_number, (size_t)(-1), false,
+                      _("too many open brackets"));
           if (extract_balanced (mlp, token_type_rbracket,
-                                null_context, null_context_list_iterator,
+                                null_context_region (),
+                                null_context_list_iterator,
                                 arglist_parser_alloc (mlp, NULL)))
             {
               arglist_parser_done (argparser, arg);
+              unref_region (inner_region);
               return true;
             }
+          bracket_nesting_depth--;
           next_context_iter = null_context_list_iterator;
           state = 0;
           continue;
@@ -1702,6 +1923,7 @@ extract_balanced (message_list_ty *mlp,
           if (delim == token_type_rbracket || delim == token_type_eof)
             {
               arglist_parser_done (argparser, arg);
+              unref_region (inner_region);
               return false;
             }
           next_context_iter = null_context_list_iterator;
@@ -1709,13 +1931,20 @@ extract_balanced (message_list_ty *mlp,
           continue;
 
         case token_type_lbrace:
+          if (++brace_nesting_depth > MAX_NESTING_DEPTH)
+            if_error (IF_SEVERITY_FATAL_ERROR,
+                      logical_file_name, line_number, (size_t)(-1), false,
+                      _("too many open braces"));
           if (extract_balanced (mlp, token_type_rbrace,
-                                null_context, null_context_list_iterator,
+                                null_context_region (),
+                                null_context_list_iterator,
                                 arglist_parser_alloc (mlp, NULL)))
             {
               arglist_parser_done (argparser, arg);
+              unref_region (inner_region);
               return true;
             }
+          brace_nesting_depth--;
           next_context_iter = null_context_list_iterator;
           state = 0;
           continue;
@@ -1724,6 +1953,7 @@ extract_balanced (message_list_ty *mlp,
           if (delim == token_type_rbrace || delim == token_type_eof)
             {
               arglist_parser_done (argparser, arg);
+              unref_region (inner_region);
               return false;
             }
           next_context_iter = null_context_list_iterator;
@@ -1732,25 +1962,68 @@ extract_balanced (message_list_ty *mlp,
 
         case token_type_string:
         case token_type_template:
+        case token_type_rtemplate:
           {
             lex_pos_ty pos;
 
             pos.file_name = logical_file_name;
             pos.line_number = token.line_number;
 
-            if (extract_all)
+            mixed_string_ty *mixed_string =
+              (token.type != token_type_rtemplate ? token.mixed_string : NULL);
+            /* For a tagged template literal, perform the tag step 1.  */
+            if ((token.type == token_type_template
+                 || token.type == token_type_rtemplate)
+                && token.template_tag != NULL)
               {
-                char *string = mixed_string_contents (token.mixed_string);
-                mixed_string_free (token.mixed_string);
-                remember_a_message (mlp, NULL, string, true, false,
-                                    inner_context, &pos,
-                                    NULL, token.comment, true);
+                const char *tag = token.template_tag;
+                string_list_ty *parts;
+
+                if (token.type == token_type_template)
+                  {
+                    parts = string_list_alloc ();
+                    string_list_append_move (parts,
+                                             mixed_string_contents (mixed_string));
+                  }
+                else /* (token.type == token_type_rtemplate) */
+                  parts = token.template_parts;
+
+                void *tag_value;
+                if (tags.table != NULL
+                    && hash_find_entry (&tags, tag, strlen (tag), &tag_value) == 0)
+                  {
+                    struct tag_definition *def = tag_value;
+
+                    /* Invoke the tag step 1 function.  */
+                    char *string = def->step1_fn (parts);
+
+                    /* Extract the string.  */
+                    remember_a_message (mlp, NULL, string, true, false,
+                                        inner_region, &pos,
+                                        NULL, token.comment, true);
+                  }
+
+                string_list_free (parts);
+
+                /* Due to the tag, the value is not a constant.  */
+                mixed_string = NULL;
+              }
+
+            if (mixed_string != NULL)
+              {
+                if (extract_all)
+                  {
+                    char *string = mixed_string_contents_free1 (mixed_string);
+                    remember_a_message (mlp, NULL, string, true, false,
+                                        inner_region, &pos,
+                                        NULL, token.comment, true);
+                  }
+                else
+                  arglist_parser_remember (argparser, arg, mixed_string,
+                                           inner_region,
+                                           pos.file_name, pos.line_number,
+                                           token.comment, true);
               }
-            else
-              arglist_parser_remember (argparser, arg, token.mixed_string,
-                                       inner_context,
-                                       pos.file_name, pos.line_number,
-                                       token.comment, true);
           }
           drop_reference (token.comment);
           next_context_iter = null_context_list_iterator;
@@ -1758,13 +2031,20 @@ extract_balanced (message_list_ty *mlp,
           continue;
 
         case token_type_xml_element_start:
+          if (++xml_element_nesting_depth > MAX_NESTING_DEPTH)
+            if_error (IF_SEVERITY_FATAL_ERROR,
+                      logical_file_name, line_number, (size_t)(-1), false,
+                      _("too many open XML elements"));
           if (extract_balanced (mlp, token_type_xml_element_end,
-                                null_context, null_context_list_iterator,
+                                null_context_region (),
+                                null_context_list_iterator,
                                 arglist_parser_alloc (mlp, NULL)))
             {
               arglist_parser_done (argparser, arg);
+              unref_region (inner_region);
               return true;
             }
+          xml_element_nesting_depth--;
           next_context_iter = null_context_list_iterator;
           state = 0;
           continue;
@@ -1773,6 +2053,7 @@ extract_balanced (message_list_ty *mlp,
           if (delim == token_type_xml_element_end || delim == token_type_eof)
             {
               arglist_parser_done (argparser, arg);
+              unref_region (inner_region);
               return false;
             }
           next_context_iter = null_context_list_iterator;
@@ -1781,11 +2062,11 @@ extract_balanced (message_list_ty *mlp,
 
         case token_type_eof:
           arglist_parser_done (argparser, arg);
+          unref_region (inner_region);
           return true;
 
         case token_type_ltemplate:
         case token_type_mtemplate:
-        case token_type_rtemplate:
         case token_type_keyword:
         case token_type_start:
         case token_type_dot:
@@ -1809,9 +2090,9 @@ extract_balanced (message_list_ty *mlp,
 
 void
 extract_javascript (FILE *f,
-                const char *real_filename, const char *logical_filename,
-                flag_context_list_table_ty *flag_table,
-                msgdomain_list_ty *mdlp)
+                    const char *real_filename, const char *logical_filename,
+                    flag_context_list_table_ty *flag_table,
+                    msgdomain_list_ty *mdlp)
 {
   message_list_ty *mlp = mdlp->item[0]->messages;
 
@@ -1846,19 +2127,20 @@ extract_javascript (FILE *f,
   phase5_pushback_length = 0;
   last_token_type = token_type_start;
 
-  template_literal_depth = 0;
-  new_brace_depth_level ();
-  xml_element_depth = 0;
-  inside_embedded_js_in_xml = false;
+  level = 0;
 
   flag_context_list_table = flag_table;
+  paren_nesting_depth = 0;
+  bracket_nesting_depth = 0;
+  brace_nesting_depth = 0;
+  xml_element_nesting_depth = 0;
 
   init_keywords ();
 
   /* Eat tokens until eof is seen.  When extract_balanced returns
      due to an unbalanced closing parenthesis, just restart it.  */
   while (!extract_balanced (mlp, token_type_eof,
-                            null_context, null_context_list_iterator,
+                            null_context_region (), null_context_list_iterator,
                             arglist_parser_alloc (mlp, NULL)))
     ;
 
